package ru.dbotthepony.kstarbound.io

import org.lwjgl.stb.STBImage
import org.lwjgl.system.MemoryUtil.memAddress
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i
import ru.dbotthepony.kvector.vector.nint.Vector4i
import java.lang.ref.Cleaner
import java.nio.ByteBuffer

private val cleaner = Cleaner.create { Thread(it, "STB Image Cleaner") }

class ImageData(val data: ByteBuffer, val width: Int, val height: Int, val amountOfChannels: Int) {
	init {
		val address = memAddress(data)
		cleaner.register(data) { STBImage.nstbi_image_free(address) }

		check(width >= 0) { "Invalid width $width" }
		check(height >= 0) { "Invalid height $height" }
		check(amountOfChannels in 1 .. 4) { "Unknown number of channels $amountOfChannels" }
	}

	operator fun get(x: Int, y: Int): Int {
		require(x in 0 until width) { "X out of bounds: $x" }
		require(y in 0 until height) { "Y out of bounds: $y" }

		val offset = y * width * amountOfChannels + x * amountOfChannels

		when (amountOfChannels) {
			4 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8) or
				data[offset + 2].toInt().shl(16) or
				data[offset + 3].toInt().shl(24)

			3 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8) or
				data[offset + 2].toInt().shl(16)

			2 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8)

			1 -> return data[offset].toInt()

			else -> throw IllegalStateException()
		}
	}

	fun worldSpaces(pos: Vector2d, spaceScan: Double, flip: Boolean): List<Vector2i> {
		when (amountOfChannels) {
			3 -> TODO()
			4 -> TODO()
			else -> throw IllegalStateException("Can not check world space taken by image with $amountOfChannels color channels")
		}
	}

	val nonEmptyRegion by lazy {
		if (amountOfChannels == 4) {
			var x0 = 0
			var y0 = 0

			search@for (y in 0 until height) {
				for (x in 0 until width) {
					if (this[x, y] and 0x000000FF != 0x0) {
						x0 = x
						y0 = y
						break@search
					}
				}
			}

			var x1 = x0
			var y1 = y0

			search@for (y in height - 1 downTo y0) {
				for (x in width - 1 downTo x0) {
					if (this[x, y] and 0x000000FF != 0x0) {
						x1 = x
						y1 = y
						break@search
					}
				}
			}

			return@lazy Vector4i(x0, y0, x1, y1)
		}

		Vector4i(0, 0, width, height)
	}
}
